@@ -1,5 +1,24 @@
+/* bug-buddy bug submitting program
+ *
+ * Copyright (C) 2007 Fernando Herrera
+ *
+ * Authors: Fernando Herrera  <fherrera@onirica.com>
+ *          Cosimo Cecchi     <cosimoc@gnome.org>
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of version 2 of the GNU General Public
+ * License as published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
+ */
 
-#include <pthread.h>
 #include <unistd.h>
 #include <cstdio>
 #include <cstdlib>
@@ -38,11 +57,22 @@ static bool    run_bug_buddy   (const gchar *appname, pid_t pid, const gchar *mi
 static void    check_if_gdb    ();
 
 static gchar *bugbuddy;
+static GLogFunc old_handler = NULL;
+
+typedef struct {
+	int start;
+	int next_free;
+	int size;
+	char **strings;
+} CircBuff;
+
+static CircBuff * log_buff = NULL;
+#define LOG_BUFFER_SIZE 15
 
 #ifdef ENABLE_GOOGLE_BREAKPAD
 static ExceptionHandler * handler;
 
-// Callback when minidump written.
+/* Callback when minidump written. */
 static bool MinidumpCallback(const char *dump_path,
                              const char *minidump_id,
                              void *context,
@@ -62,6 +92,107 @@ static bool MinidumpCallback(const char *dump_path,
 }
 #endif
 
+static void
+circ_buff_free (CircBuff *buff)
+{
+	g_strfreev (buff->strings);
+	g_free (buff);
+
+	buff = NULL;
+}
+
+static CircBuff *
+circ_buff_new (int size)
+{
+	CircBuff *retval;
+
+	retval = g_new0 (CircBuff, 1);
+	retval->strings = g_new0 (char *, size + 1);
+	retval->strings[size + 1] = NULL;
+	retval->size = size;
+
+	return retval;
+}
+
+static void
+circ_add (CircBuff *buff, char * str)
+{
+	if (buff->strings[buff->next_free]) {
+		g_free (buff->strings[buff->next_free]);
+		buff->start = (buff->start + 1) % buff->size;
+	}
+	
+	buff->strings[buff->next_free] = str;
+	buff->next_free = (buff->next_free + 1) % buff->size;
+}
+
+static char *
+circ_buff_to_file (CircBuff *buff)
+{
+	int tmp_fd, i;
+	char *actual_name = NULL;
+	GError *error = NULL;
+	const char *header;
+
+	tmp_fd = g_file_open_tmp ("bug-buddy-XXXXXX", &actual_name, &error);
+
+	if (error) {
+		g_warning ("Unable to create the warnings temp file: %s", error->message);
+		g_error_free (error);
+		return NULL;
+	}
+
+	header = "\n\n---- Critical and fatal warnings logged during execution ----\n\n";
+	write (tmp_fd, header, strlen (header));
+
+	i = buff->start;
+
+	while (buff->strings[i] != NULL) {
+		/* write to file */
+		write (tmp_fd, buff->strings[i], strlen (buff->strings[i]));
+
+		/* increase and check for overflow */
+		i = (i + 1) % buff->size;
+
+		if (i == buff->start)
+			break;
+	}
+
+	close (tmp_fd);
+
+	return actual_name;
+}
+
+static void
+bug_buddy_log_handler (const gchar *log_domain,
+		       GLogLevelFlags log_level,
+		       const gchar *message,
+		       gpointer user_data)
+{
+	/* forward the message to the real handler and use the default GLib
+	 * handler if that's NULL (should never happen, but who knows).
+	 */
+	if (old_handler) {
+		old_handler (log_domain, log_level, message, user_data);
+	} else {
+		g_log_default_handler (log_domain, log_level, message, user_data);
+	}
+
+	if (log_level & 
+	    (G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL | G_LOG_FLAG_FATAL)) 
+	{
+		/* log the message into memory */
+		if (!log_buff) {
+			log_buff = circ_buff_new (LOG_BUFFER_SIZE);
+		}
+
+		/* if the log domain is NULL, assume it's a message coming from the app,
+		 * see the docs for G_LOG_DOMAIN.
+		 */
+		circ_add (log_buff, g_strdup_printf ("** %s **: %s \n", log_domain ? log_domain : g_get_prgname (), message));
+	}
+}
+
 static void
 bugbuddy_segv_handle(int signum)
 {
@@ -119,11 +250,11 @@ find_in_debug_path (const char *filename, const char *debug_filename)
 	}
 
 	for (i = 0; i < N_TRIES; i++)
-		g_free (tries [i]);
+		g_free (tries [N_TRIES]);
 
 	return result;
 }
-	
+
 
 static gboolean
 elf_has_debug_symbols (const char *filename)
@@ -201,29 +332,36 @@ release_grabs (void)
 static bool
 run_bug_buddy (const gchar *appname, pid_t pid, const gchar *minidump_path)
 {
-	gchar *exec_str;
 	gboolean res;
+	char *warning_file, *exec_str;
+	GString *args_str;
 	GError *error = NULL;
 
-	if (pid != 0) 
-		exec_str = g_strdup_printf("bug-buddy "
-				   	"--appname=\"%s\" "
-				  	 "--pid=%d",
-				   	appname,(int) pid);
-	else if (minidump_path != NULL)
-		exec_str = g_strdup_printf("bug-buddy "
-				   	"--appname=\"%s\" "
-				  	 "--minidump=%s",
-				   	appname, minidump_path);
-	else
+	if (pid == 0 && minidump_path == NULL)
 		return false;
 
+	warning_file = circ_buff_to_file (log_buff);
+	circ_buff_free (log_buff);
 
+	args_str = g_string_new ("bug-buddy ");
+	g_string_append_printf (args_str, "--appname=\"%s\" ", appname);
+
+	if (warning_file) {
+		g_string_append_printf (args_str, "--include=\"%s\" --unlink-tempfile ", warning_file);
+		g_free (warning_file);
+	}
+
+	if (pid != 0)
+		g_string_append_printf (args_str,  "--pid=%d",(int) pid);
+	else if (minidump_path != NULL)
+		g_string_append_printf (args_str, "--minidump=%s", minidump_path);
+
+	exec_str = g_string_free (args_str, FALSE);
 	res = g_spawn_command_line_sync (exec_str, NULL, NULL,
 					 NULL, &error);
-	g_free(exec_str);
+	g_free (exec_str);
 	if (!res) {
-		g_warning("Couldn't run bug-buddy\n");
+		g_warning ("Couldn't run bug-buddy\n");
 		return false;
 	}
 
@@ -260,9 +398,9 @@ run_gdb (const gchar *appname, pid_t pid)
 static void
 check_if_gdb ()
 {
-	char mypath[255];
+	char *mypath;
 	gchar *gdb;
-	bool has_debug_symbols;
+	bool has_debug_symbols = false;
 	char *filename;
 	gchar *appname;
 	pid_t pid;
@@ -284,9 +422,11 @@ check_if_gdb ()
 		_exit(0);
 	}
 	
-	memset(mypath, 0, sizeof(mypath));
-	readlink ("/proc/self/exe", mypath, sizeof(mypath));
-	has_debug_symbols = elf_has_debug_symbols (mypath);
+	mypath = g_file_read_link ("/proc/self/exe", NULL);
+	if (mypath)
+		has_debug_symbols = elf_has_debug_symbols (mypath);
+
+	g_free (mypath);
 
 	if (bugbuddy && gdb && has_debug_symbols) {
 		res = run_bug_buddy (appname, pid, NULL);
@@ -335,6 +475,8 @@ gtk_module_init (int *argc, char** argv[])
                 sigaction(SIGBUS, NULL, &old_action);
 		if (old_action.sa_handler == SIG_DFL)
                 	sigaction(SIGBUS, setptr, NULL);
+
+		old_handler = g_log_set_default_handler (bug_buddy_log_handler, NULL);
 	}
 	return 0;
 }
